from __future__ import annotations

import http.server
import os
import shutil
import subprocess
from http.server import HTTPServer
from pathlib import Path
from threading import Thread
from typing import Optional, Union, List

from .view import View
from ...util.server import AbstractServer
from ...util import tensorneko_util_path


class Server(AbstractServer):
    """
    The http server for display watched components.

    Args:
        view (:class:`~tensorneko.visualization.web.view.View` | ``str``):
            The :class:`~tensorneko.visualization.watcher.view.View` object or the name of
            :class:`~tensorneko.visualization.watcher.view.View` for display.

        port (``int``, optional): The port of the http server. Default 8000.

    Attributes:
        view_name (``str``): The name of the :class:`~tensorneko.visualization.watcher.view.View`.
        process (:class:``subprocess.Popen``): The process of the server.
        save_dir (``str``): The directory to save the web files.

    Examples::

        # start a server
        var = tensorneko.visualization.watcher.Variable("x", 5)
        view = tensorneko.visualization.watcher.View("view1")
        view.add(var)
        server = tensorneko.visualization.watcher.Server(view)
        server.start()

        # stop the server
        server.stop()

        # start the server on blocking
        server.start_blocking()

    """
    servers: List[Server] = []

    def __init__(self, view: Union[View, str], port: int = 8000, save_dir: str = None):
        if type(view) == View:
            self.view_name = view.name
        elif type(view) == str:
            self.view_name = view
        else:
            raise TypeError("The view type should be View or str")

        self.save_dir = save_dir or os.path.join("watcher", self.view_name)
        self.port = port
        self.process: Optional[Thread] = None
        self.httpd: Optional[HTTPServer] = None
        server_dir = os.path.join("watcher", self.view_name)

        class HTTPHandler(http.server.SimpleHTTPRequestHandler):

            def __init__(self, *args, **kwargs):
                super().__init__(*args, directory=server_dir, **kwargs)

            def log_message(self, format: str, *args) -> None:
                return

        self.Handler = HTTPHandler
        self.Handler.extensions_map[".js"] = "application/javascript"
        self.Handler.extensions_map[".mjs"] = "application/javascript"
        http.server.HTTPServer.allow_reuse_address = True
        http.server.HTTPServer.allow_reuse_port = True
        Server.servers.append(self)

    def start(self) -> None:
        """
        Start or restart the server.
        """
        # stop the old server
        if self.process is not None:
            self.stop()
        # copy files to the view directory
        self._prepare()
        # run python http server
        self._run()

    def start_blocking(self) -> None:
        """
        Start or restart the server on blocking.
        """
        # stop the old server
        if self.process is not None:
            self.stop()
        # copy files to the view directory
        self._prepare()
        # run python http server
        self._run_blocking()

    def _prepare(self) -> None:
        source_path = os.path.join(tensorneko_util_path, "visualization", "watcher", "web", "dist")
        target_path = self.save_dir
        if not os.path.exists(target_path):
            Path(target_path).mkdir(parents=True, exist_ok=True)

        target_path_assets = os.path.join(target_path, "assets")
        if os.path.exists(target_path_assets):
            shutil.rmtree(target_path_assets)

        target_path_html = os.path.join(target_path, "index.html")
        if os.path.exists(target_path_html):
            os.remove(target_path_html)

        shutil.copy(os.path.join(source_path, "index.html"), target_path_html)
        shutil.copytree(os.path.join(source_path, "assets"), target_path_assets)

    def _run(self) -> None:
        self.httpd = http.server.HTTPServer(("", self.port), self.Handler)
        self.process = Thread(target=self.httpd.serve_forever)
        self.process.start()
        print(self.server_start_string)

    def _run_blocking(self):
        with http.server.HTTPServer(("", self.port), self.Handler) as httpd:
            print(self.server_start_string)
            httpd.serve_forever()

    def stop(self) -> None:
        """
        Stop the server.
        """
        if self.process is not None:
            self.httpd.shutdown()
            self.process.join()
            print("Server stopped")
            self.process = None
        else:
            print("Server is stopped.")

    @classmethod
    def stop_all(cls) -> None:
        """
        Stop all servers.
        """
        for server in cls.servers:
            server.stop()

    @property
    def server_start_string(self) -> str:
        return f"Server started at http://127.0.0.1:{self.port}/, view \"{self.view_name}\"."



from typing import Optional

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from numpy import ndarray


def plot_image(image: ndarray, title: Optional[str] = None, figure: Optional[Figure] = None, **kwargs) -> Figure:
    fig = figure or plt.figure()
    ax = fig.add_subplot()
    ax.imshow(image, **kwargs)
    ax.set_axis_off()
    if title is not None:
        ax.set_title(title)
    return fig



class TextReader:
    """TextReader for reading text file"""

    @staticmethod
    def of_plain(path: str, encoding: str = "UTF-8") -> str:
        """
        Read texts of a file.

        Args:
            path (``str``): Text file path.
            encoding (``str``, optional): File encoding. Default "UTF-8".

        Returns:
            ``str``: The texts of given file.
        """
        with open(path, "r", encoding=encoding) as file:
            text = file.read()
        return text

    of = of_plain

    def __new__(cls, path: str, encoding: str = "UTF-8") -> str:
        """Alias of :meth:`~TextReader.of"""
        return cls.of(path, encoding)



from . import type
from .average_meter import AverageMeter
from .dispatcher import dispatch
from .fp import __, F, _, Stream, return_option, Option, Monad, Eval, Seq, AbstractSeq, curry
from .func import generate_inf_seq, compose, listdir, with_printed, ifelse, dict_add, as_list, identity, list_to_dict, \
    get_tensorneko_util_path, circular_pad
from .dispatched_func import sparse2binary, binary2sparse
from .ref import ref
from .timer import Timer
from .eventbus import Event, EventBus, EventHandler, subscribe, subscribe_async, subscribe_process, subscribe_thread
from .singleton import Singleton

tensorneko_util_path = get_tensorneko_util_path()


__all__ = [
    "generate_inf_seq",
    "compose",
    "listdir",
    "with_printed",
    "ifelse",
    "dict_add",
    "as_list",
    "list_to_dict",
    "type",
    "__",
    "_",
    "F",
    "Stream",
    "return_option",
    "Option",
    "Monad",
    "Eval",
    "dispatch",
    "AverageMeter",
    "curry",
    "AbstractSeq",
    "Seq",
    "tensorneko_util_path",
    "sparse2binary",
    "binary2sparse",
    "ref",
    "Timer",
    "Event",
    "EventBus",
    "subscribe",
    "subscribe_async",
    "subscribe_process",
    "subscribe_thread",
    "Singleton",
    "circular_pad",
]



import os
from functools import reduce
from os.path import dirname, abspath
from typing import Callable, List, Dict, Iterable, Sequence, Any

from .fp import F, _, Stream
from .type import T, R


def identity(*args, **kwargs):
    """
    Pickle-friendly identity function.
    """
    assert kwargs == {}, "identity function doesn't accept keyword arguments"
    if len(args) == 1:
        return args[0]
    else:
        return args


def generate_inf_seq(items: Iterable[Any]) -> Stream:
    """
    Generate an infinity late-evaluate sequence.

    Args:
        items [``Iterable[Any]``]: Original items.

    Returns:
        :class:`~fp.stream.Stream`: The infinity sequence.

    Examples::

        >>> seq = generate_inf_seq(["a", 1, ["e0", "e1"]])
        >>> list(seq[:10])
        ['a', 1, ['e0', 'e1'], 'a', 1, ['e0', 'e1'], 'a', 1, ['e0', 'e1'], 'a']

    Note:
        The repeated reference type objects are the same object.

        >>> a_list = ["a", "b"]
        >>> seq = list(generate_inf_seq([a_list])[:3])
        >>> seq
        [['a', 'b'], ['a', 'b'], ['a', 'b']]
        >>> a_list.append("c")
        >>> seq
        [['a', 'b', 'c'], ['a', 'b', 'c'], ['a', 'b', 'c']]

    """
    s = Stream()
    return s << items << map(_, s)


def compose(fs: Sequence[Callable]) -> F:
    """
    Compose functions as a pipeline function.

    Args:
        fs (``Sequence[Callable]``): The functions input for composition.

    Returns:
        :class:`~fp.func.F`: The composed output function.

    Examples::

        >>> f = lambda x: x + 1
        >>> g = lambda x: x * 2
        >>> h = lambda x: x ** 2
        >>> x = 1
        >>> h(g(f(x))) == compose([f, g, h])(x)
        True

    """
    return reduce(_ >> _, fs, F())


def listdir(path: str, filter_func: Callable[[str], bool] = lambda arg: True) -> List[str]:
    """
    The full path version of :func:`os.listdir`.

    Get the all listdir result with input path.

    Args:
        path (``str``): The path for listdir.
        filter_func (``(str) -> bool``, optional): The filter function to filter the file/directory name in ``listdir``.

    Returns:
        ``List[str]``: listdir result with input path.

    Examples::

        >>> os.listdir("tensorneko/util")[:3]  # python os library
        ['configuration.py', 'enum.py', 'string_getter.py']
        >>> listdir("tensorneko/util")[:3]  # tensorneko
        ['tensorneko/util/configuration.py',
         'tensorneko/util/enum.py',
         'tensorneko/util/string_getter.py']

    """
    files = filter(filter_func, os.listdir(path))
    return list(map(F(os.path.join, path), files))


def with_printed(x: Any, func: Callable = identity) -> Any:
    """
    An identity function but with printed to console with some transform.

    Args:
        x (``Any``): Input.
        func (``Callable``, optional): A function used to apply the input for printing.

    Returns:
        ``Any``: Identity output.

    Examples::

        >>> x = torch.tensor([1.5, 2.5, 3.5])
        >>> y = with_printed(x, lambda tensor: tensor.dtype)
        torch.float32
        >>> x == y
        tensor([True, True, True])

    """
    print(func(x))
    return x


def ifelse(predicate: Callable[[T], bool], func_true: Callable[[T], R], func_false: Callable[[T], R] = identity
) -> Callable[[T], R]:
    """
    A function composition util for if-else control flow.

    Args:
        predicate (``(T) -> bool``): A predicate produce a bool.
        func_true (``(T) -> R``): If the bool from predicate is True, return this function.
        func_false (``(T) -> R``): If the bool from predicate is False, return this function.

    Returns:
        ``(T) -> R``: Composed function with if-else control flow.
    """

    def wrapper(*args, **kwargs):
        if predicate(*args, **kwargs):
            return func_true(*args, **kwargs)
        else:
            return func_false(*args, **kwargs)

    return wrapper


def dict_add(*dicts: dict) -> dict:
    """
    Merge multiple dictionaries.

    Args:
        *dicts: dictionaries to merge.

    Returns:
        ``dict``: The merged dictionary.
    """
    new_dict = {}
    for each in dicts:
        new_dict.update(each)
    return new_dict


def as_list(*args, **kwargs) -> list:
    """
    Returns:
        ``list``: The list of args.

    Examples::

        >>> as_list(1, 2, 3, 4, key=5)
        [1, 2, 3, 4, 5]
    """
    return list(args) + list(kwargs.values())


def list_to_dict(l: List[T], key: Callable[[T], Any]) -> Dict[Any, T]:
    """
    Convert the list as a dictionary by given key.
    Args:
        l (``List[T]``): Input list.
        key (``(T) -> Any``): The key getter function.

    Returns:
        ``Dict[Any, T]``: The output dictionary.
    """
    return {key(x): x for x in l}


def get_tensorneko_util_path() -> str:
    """
    Get the `tensorneko_util` library root path

    Returns:
        ``str``: The root path of `tensorneko`
    """
    return dirname(dirname(abspath(__file__)))


def circular_pad(x: List, target: int) -> List:
    """
    Circular padding a list to the target length.

    Args:
        x(``List``): Input list.
        target(``int``): Target length.

    Returns:
        ``List``: The padded list.
    """
    if len(x) == target:
        return x
    elif len(x) > target:
        return x[:target]
    elif 2 * len(x) > target:
        return x + x[:target - len(x)]
    else:
        return circular_pad(x + x, target)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Callable, TypeVar

from ...type import T_CO as T, R


class Monad(Generic[T], ABC):

    # method `pure` is equivalent to the constructor

    @abstractmethod
    def flat_map(self, f: Callable[[T], Monad[R]]) -> Monad[R]:
        ...

    @abstractmethod
    def map(self, f: Callable[[T], R]) -> Monad[R]:
        ...

    @abstractmethod
    def flatten(self: Monad[Monad[T]]) -> Monad[T]:
        ...



from typing import Optional

import numpy as np
from einops import rearrange

from ...backend.visual_lib import VisualLib
from ...util.type import T_ARRAY
from .._default_backends import _default_image_io_backend


class ImageReader:
    """ImageReader for reading images as :class:`~torch.Tensor`"""

    @classmethod
    def of(cls, path: str, channel_first: bool = False, backend: Optional[VisualLib] = None) -> T_ARRAY:
        """
        Read image tensor of given file.

        Args:
            path (``str``): Path of the JPEG or PNG image.
            channel_first (``bool``, optional): Get image dimension (H, W, C) if False or (C, H, W) if True.
                Default: False.
            backend (:class:`~tensorneko_util.backend.visual_lib.VisualLib`, optional): The backend library for saving.
                Default: "opencv" if installed else "matplotlib".

        Returns:
            :class:`~numpy.ndarray` | :class:`~torch.Tensor`: A float tensor of image, with value range of 0. to 1.
        """
        backend = backend or _default_image_io_backend()

        if backend == VisualLib.OPENCV:
            if not VisualLib.opencv_available():
                raise ValueError("OpenCV is not installed.")
            import cv2
            img = cv2.imread(path).astype(np.float32) / 255.0
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        elif backend == VisualLib.MATPLOTLIB:
            if not VisualLib.matplotlib_available():
                raise ValueError("Matplotlib is not installed.")
            import matplotlib.pyplot as plt
            img = plt.imread(path).astype(np.float32) / 255.0
        elif backend == VisualLib.PYTORCH:
            if not VisualLib.pytorch_available():
                raise ValueError("Torchvision is not installed.")
            import torchvision
            img = torchvision.io.read_image(path).float() / 255.0  # (C, H, W)
        elif backend == VisualLib.PIL:
            if not VisualLib.pil_available():
                raise ValueError("PIL is not installed.")
            import PIL.Image
            img = np.asarray(PIL.Image.open(path)).astype(np.float32) / 255.0
        else:
            raise ValueError("Unknown image reader backend: {}".format(backend))

        if channel_first:
            if backend != VisualLib.PYTORCH:
                img = rearrange(img, 'H W C -> C H W')
        else:
            if backend == VisualLib.PYTORCH:
                img = rearrange(img, 'C H W -> H W C')

        return img

    def __new__(cls, path: str, channel_first: bool = True, backend: Optional[VisualLib] = None) -> T_ARRAY:
        """Alias of :meth:`~ImageReader.of`"""
        return cls.of(path, channel_first, backend)



import os.path
import unittest

import tensorneko


class TestLibraryInfo(unittest.TestCase):

    def test_version(self):
        version = tensorneko.io.read.text.of("version.txt")

        tensorneko.io.write.text.to(os.path.join("src", "tensorneko", "version.txt"), version)
        self.assertEqual(tensorneko.__version__, version)


if __name__ == '__main__':
    unittest.main()



from typing import Union, Optional, Tuple

from numpy import ndarray

from .abstract_face_detector import AbstractFaceDetector
from ..crop import crop_with_padding
from ..image import rgb2gray
from ...backend import VisualLib


class OpencvFaceDetector(AbstractFaceDetector):
    """
    The face detector implemented in OpenCV.
    The library OpenCV and opencv-contrib-python are required.

    Examples::

        from tensorneko_util.preprocess.face_detector import FaceXZooFaceDetector

        detector = OpencvFaceDetector()

        # process image, save to opencv_face_detector_out.jpg
        detector.process_image("test/resource/image_sample/1.183.jpg",
            "test/resource/temp_output/opencv_face_detector_out.jpg")

        # process video, save to opencv_face_detector_out.mp4
        detector.process_video("test/resource/video_sample/sample.mp4",
            "test/resource/temp_output/opencv_face_detector_out.mp4")

    """

    def __init__(self):
        try:
            import cv2
        except ImportError:
            raise ImportError("OpenCV and opencv-contrib-python is not installed. Please install it first.")

        try:
            import cv2.data
        except ImportError:
            raise ImportError("opencv-contrib-python is not installed. Please install it first.")

        self.cv2 = cv2
        self.face_cascade = cv2.CascadeClassifier(cv2.data.haarcascades + "haarcascade_frontalface_default.xml")

    def detect_face(self, image: ndarray, *args, **kwargs):
        # img should be gray scale
        if len(image.shape) == 3:
            image = rgb2gray((image * 255).astype(int), backend=VisualLib.OPENCV)

        faces = self.face_cascade.detectMultiScale(image, *args, **kwargs)
        return faces

    def process_image(self, image_path: str, out_path: str, max_faces=1, margin=0, *args, **kwargs):
        if max_faces > 1:
            raise NotImplementedError("Multiple faces are not supported yet.")

        img = self.cv2.imread(image_path)
        gray = self.cv2.cvtColor(img, self.cv2.COLOR_BGR2GRAY)
        faces = self.detect_face(gray, *args, **kwargs)
        for (x, y, w, h) in faces[:max_faces]:
            cropped = crop_with_padding(img, x - margin, x + w + margin, y - margin, y + h + margin)
            self.cv2.imwrite(out_path, cropped)

    def _infer_frame_size(self, video_path: str, margin: int = 0) -> Tuple[int, int]:
        # get first face crop in the video
        video = self.cv2.VideoCapture(video_path)
        while True:
            ret, frame = video.read()
            if not ret:
                break
            gray = self.cv2.cvtColor(frame, self.cv2.COLOR_BGR2GRAY)
            faces = self.detect_face(gray)
            if len(faces) > 0:
                x, y, w, h = faces[0]
                side = max(w, h)
                video.release()
                return side + 2 * margin, side + 2 * margin

    def process_video(self, video_path: str, out_path: str, frame_size: Optional[Union[int, Tuple[int, int]]] = None,
        margin=0, fourcc="mp4v", *args, **kwargs
    ):
        video = self.cv2.VideoCapture(video_path)
        if not video.isOpened():
            raise IOError("Cannot open video file: " + video_path)

        # infer frame size
        if frame_size is None:
            frame_size = self._infer_frame_size(video_path, margin)

        if type(frame_size) is int:
            frame_size = frame_size, frame_size

        fps = video.get(self.cv2.CAP_PROP_FPS)
        writer = self.cv2.VideoWriter(out_path, self.cv2.VideoWriter_fourcc(*fourcc), fps, frame_size)
        x1, y1, x2, y2 = 0, 0, 1, 1
        while True:
            ret, frame = video.read()
            if not ret:
                break
            gray = self.cv2.cvtColor(frame, self.cv2.COLOR_BGR2GRAY)
            faces = self.detect_face(gray, *args, **kwargs)
            if len(faces) > 0:
                x, y, w, h = faces[0]
                side = max(w, h)
                center = (x + w // 2, y + h // 2)
                x1, y1 = center[0] - side // 2, center[1] - side // 2
                x2, y2 = center[0] + side // 2, center[1] + side // 2

            cropped = crop_with_padding(frame, x1 - margin, x2 + margin, y1 - margin, y2 + margin)
            resized = self.cv2.resize(cropped, frame_size)
            writer.write(resized)
        video.release()
        writer.release()



from dataclasses import dataclass

from ...util.type import T_ARRAY


@dataclass
class AudioData:
    """
    Output audio data for :class:`~.audio_reader.AudioReader`.

    Attributes:
        audio (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): audio float tensor of (C, T)
        sample_rate (``int``): sample rate of the audio

    """
    audio: T_ARRAY
    sample_rate: int

    def __iter__(self):
        return iter((self.audio, self.sample_rate))



import glob
import os.path
from abc import ABC, abstractmethod

from numpy import ndarray


class AbstractFaceDetector(ABC):

    @abstractmethod
    def detect_face(self, image: ndarray, *args, **kwargs):
        pass

    @abstractmethod
    def process_image(self, image_path: str, out_path: str, *args, **kwargs):
        pass

    @abstractmethod
    def process_video(self, video_path: str, out_path: str, *args, **kwargs):
        pass

    def process_image_dir(self, image_dir: str, out_dir: str, pattern="*.jpg", *args, **kwargs):
        all_images = glob.glob(os.path.join(image_dir, pattern), root_dir=image_dir)
        for image_path in all_images:
            out_path = os.path.join(out_dir, image_path)
            self.process_image(image_path, out_path, *args, **kwargs)

    def process_video_dir(self, video_dir: str, out_dir: str, pattern="*.mp4", *args, **kwargs):
        all_videos = glob.glob(os.path.join(video_dir, pattern), root_dir=video_dir)
        for video_path in all_videos:
            out_path = os.path.join(out_dir, video_path)
            self.process_video(video_path, out_path, *args, **kwargs)



from __future__ import annotations

import itertools
from functools import reduce
from typing import Collection, List, Callable, Union, Iterator, Iterable, Optional, Any

from .abstract_seq import AbstractSeq
from ...type import T, R
from ....backend.parallel import ExecutorPool, ParallelType


def _identity(x: T) -> T:
    return x


class Seq(AbstractSeq[T], Collection[T]):

    def __init__(self, items: Iterable[T] = iter(())):
        if isinstance(items, Seq):
            self._items = items.to_list()
        else:
            self._items = list(items)

    @staticmethod
    def of(*item: T) -> Seq[T]:
        return Seq(item)

    @staticmethod
    def from_seq(seq: Seq[T]) -> Seq[T]:
        return Seq(seq)

    def __lshift__(self, right_iter: Iterable[T]) -> Seq[T]:
        return Seq(self._items + list(right_iter))

    def __getitem__(self, item: Union[int, slice]) -> Union[T, AbstractSeq[T]]:
        if isinstance(item, int):
            return self._items[item]
        elif isinstance(item, slice):
            return Seq(self._items[item])
        else:
            raise TypeError(f"{type(item)} is not supported")

    def __iter__(self) -> Iterator[T]:
        return iter(self._items)

    def __len__(self) -> int:
        return len(self._items)

    def __contains__(self, __x: object) -> bool:
        return self._items.__contains__(__x)

    def map(self, f: Callable[[T], R], progress_bar: bool = False, parallel_type: Optional[ParallelType] = None,
        **tqdm_args
    ) -> Seq[R]:
        if parallel_type is None:
            items = self._items if not progress_bar else self._tqdm(self._items, **tqdm_args)
            return Seq(map(f, items))
        else:
            futures = []

            if f.__name__ == "<lambda>":
                raise NotImplementedError("lambda function is not supported yet")

            for item in self._items:
                futures.append(ExecutorPool.submit(f, item, parallel_type=parallel_type))

            futures = self._tqdm(futures, **tqdm_args) if progress_bar else futures
            return Seq(map(lambda future: future.result(), futures))

    def for_each(self, f: Callable[[T], None], progress_bar: bool = False, parallel_type: Optional[ParallelType] = None,
        **tqdm_args
    ) -> None:
        if parallel_type is None:
            items = self._items if not progress_bar else self._tqdm(self._items, **tqdm_args)
            for item in items:
                f(item)
        else:
            self.map(f, progress_bar, parallel_type, **tqdm_args)

    def with_for_each(self, f: Callable[[T], None], progress_bar: bool = False,
        parallel_type: Optional[ParallelType] = None, **tqdm_args
    ) -> Seq[T]:
        self.for_each(f, progress_bar, parallel_type, **tqdm_args)
        return self

    def filter(self, f: Callable[[T], bool]) -> Seq[T]:
        return Seq(filter(f, self._items))

    def reduce(self, f: Callable[[T, T], T]) -> T:
        return reduce(f, self._items)

    def flatten(self) -> Seq[T]:
        new_items = []
        for item in self._items:
            if isinstance(item, Seq):
                new_items.extend(item.to_list())
            else:
                new_items.append(item)
        return Seq(new_items)

    def flat_map(self, f: Callable[[T], AbstractSeq[R]], progress_bar: bool = False) -> Seq[R]:
        return self.map(f, progress_bar).flatten()

    def sort(self, key: Callable[[T], Any] = _identity, reverse: bool = False) -> Seq[T]:
        return Seq(sorted(self._items, key=key, reverse=reverse))

    def take(self, n: int) -> Seq[T]:
        return self[:n]

    def skip(self, n: int) -> Seq[T]:
        return self[n:]

    @property
    def head(self) -> T:
        return self._items[0]

    @property
    def tail(self) -> Seq[T]:
        return self[1:]

    def repeat(self, n: int) -> Seq[T]:
        self._items = [item for item in itertools.repeat(self._items, n)]
        return self

    def to_list(self) -> List[T]:
        return self._items

    def __str__(self):
        return f"Seq({', '.join(map(str, self._items))})"

    @property
    def length(self):
        return len(self._items)

    @property
    def size(self):
        return self.length

    __repr__ = __str__



from enum import Enum


class Colors(Enum):
    GREEN = "#C3E88D"
    YELLOW = "#FFCB6B"
    RED = "#F07178"
    BLUE = "#82AAFF"
    PURPLE = "#C792EA"
    ORANGE = "#F78C6C"
    AZURE = "#467CDA"
    SKY = "#89DDFF"
    BLACK = "#2B2B2B"
    GRAY = "#616161"
    WHITE = "#EEFFFF"
    BLANK = "#FFFFFF"


class ContinuousColors(Enum):
    RED = "#d11141",
    YELLOW = "#ffc425",
    GREEN = "#00b159",
    BLACK = "#2B2B2B",
    WHITE = "#EEFFFF"



from __future__ import annotations

import json
import os
from dataclasses import dataclass, field
from typing import Union, List

from . import Component


@dataclass
class View:
    """
    The view for a server for display.

    Args:
        name (``str``): The name of the view.

    Attributes:
        components (``List[Component]``): The :class:`~tensorneko.visualization.watcher.component.Component` objects in
            the view.

    Examples::

        # create components
        var = tensorneko.visualization.watcher.Variable("x", 5)
        pb = tensorneko.visualization.watcher.ProgressBar("process", 0, 10000)

        # create a view and add components
        view = tensorneko.visualization.watcher.View("view1")
        view.add(var).add(pb)

        # components also support read by multiple views
        view2 = tensorneko.visualization.watcher.View("view2")
        view2.add(var)

    """
    name: str
    components: List[Component] = field(init=False, default_factory=list)

    def update(self) -> None:
        """Save the state to the json file."""
        return self._save_json()

    def _save_json(self) -> None:
        if not os.path.exists("watcher"):
            os.mkdir("watcher")

        view_path = os.path.join("watcher", self.name)
        if not os.path.exists(view_path):
            os.mkdir(view_path)

        with open(os.path.join(view_path, "data.json"), "w", encoding="UTF-8") as file:
            file.write(json.dumps({
                "view": self.name,
                "data": list(map(lambda comp: comp.to_dict(), self.components))
            }))

    def add(self, *components: Component) -> View:
        """
        Add the new component to the view.

        Args:
            components (:class:`~tensorneko.visualization.web.component.Component`): The component for adding.

        Returns:
            :class:`~tensorneko.visualization.web.view.View`: The self view object.
        """
        for component in components:
            component.views.append(self)
            self.components.append(component)
            component.update()
            self.update()
        return self

    def add_all(self) -> View:
        """
        Add all defined components.

        Returns:
            :class:`~tensorneko.visualization.web.view.View`: The self view object.
        """
        return self.add(*Component.components.values())

    def remove(self, component: Union[Component, str]) -> View:
        """
        Remove the component in the view.

        Args:
            component (:class:`~tensorneko.visualization.web.component.Component` | ``str``):
                The component for removing.

        Returns:
            :class:`~tensorneko.visualization.web.view.View`: The self view object.
        """
        if type(component) is Component and component in self.components:
            self.components.remove(component)
        elif type(component) is str:
            for i in range(len(self.components)):
                if self.components[i].name == component:
                    self.components[i].views.remove(self)
                    del self.components[i]
                    break
            else:
                print("Component remove failed")
        else:
            print("Component remove failed")

        self.update()
        return self

    def __getitem__(self, component_name: str) -> Component:
        """
        Get the component by name.

        Args:
            component_name (``str``): The name of the component.

        Returns:
            :class:`~tensorneko.visualization.web.component.Component`: The component object.
        """
        for component in self.components:
            if component.name == component_name:
                return component
        raise KeyError(f"Component {component_name} not found")



from typing import Any

import yaml


class YamlWriter:

    @classmethod
    def to(cls, path: str, obj: Any) -> None:
        """
        Write a Python object to a YAML file.

        Args:
            path (``str``): YAML file path.
            obj (``Any``): The Python object to write.
        """
        with open(path, "w") as f:
            yaml.dump(obj, f)

    def __new__(cls, path: str, obj: Any) -> None:
        """Alias of :meth:`~tensorneko_util.io.yaml.yaml_writer.YamlWriter.to`."""
        cls.to(path, obj)
        return None



from typing import Type, Dict

from .type import T


class Singleton:
    """
    Singleton decorator. It is used to define the class as a singleton.

    Note: The constructor should not include any arguments.

    Example::

        from tensorneko.util import Singleton

        @Singleton
        class MyObject:
            def __init__(self):
                self.value = 0

            def add(self, value):
                self.value += value
                return self.value


        print(MyObject.value)  # 0
        MyObject.add(1)
        print(MyObject.value)  # 1

    """

    all_instances: Dict[Type, object] = {}

    def __new__(cls, clazz: Type[T]) -> T:
        cls.all_instances[clazz] = clazz()
        return cls.all_instances[clazz]



import json
from typing import Union


class JsonWriter:

    @classmethod
    def to(cls, path: str, obj: Union[dict, list, object], encoding: str = "UTF-8") -> None:
        """
        Save as Json file from a dictionary, list or json_dict.

        Args:
            path (``str``): The path of output file.
            obj (``dict`` | ``list`` | ``object``): The json data which need to be used for output. The type should be
                ``dict``, ``list`` or the ``object`` decorated by :func:`~tensorneko.io.text.text_reader.json_data`.
            encoding (``str``, optional): Python file IO encoding parameter. Default: "UTF-8".
        """
        if type(obj) is dict:
            with open(path, "w", encoding=encoding) as file:
                file.write(json.dumps(obj))
        elif "is_json_data" in type(obj).__dict__ and type(obj).is_json_data:
            with open(path, "w", encoding=encoding) as file:
                file.write(json.dumps(obj.to_dict()))
        elif type(obj) is list:
            if len(obj) == 0 or type(obj[0]) in (dict, list):
                with open(path, "w", encoding=encoding) as file:
                    file.write(json.dumps(obj))
            elif "is_json_data" in type(obj[0]).__dict__ and type(obj[0]).is_json_data:
                obj = [each.to_dict() for each in obj]
                with open(path, "w", encoding=encoding) as file:
                    file.write(json.dumps(obj))
        else:
            raise TypeError("Not implemented type. Only support dict, list, json_data.")

    def __new__(cls, path: str, obj: Union[dict, list, object], encoding: str = "UTF-8") -> None:
        """Alias of :meth:`~tensorneko.io.json.json_writer.JsonWriter.to`."""
        cls.to(path, obj, encoding)



import time
from functools import wraps
from typing import Callable, Optional


class Timer:
    """
    A timer to count the time of a function or a block of code.

    Examples::

        from tensorneko.util import Timer
        import time

        # use as a context manager with single time
        with Timer():
            time.sleep(1)

        # use as a context manager with multiple segments
        with Timer() as t:
            time.sleep(1)
            t.time("sleep A")
            time.sleep(1)
            t.time("sleep B")
            time.sleep(1)

        # use as a decorator
        @Timer()
        def f():
            time.sleep(1)
            print("f")

    """

    def __init__(self):
        self.times = []
        self.names = []
        self.total_time = None

    def __enter__(self):
        self.times.append(time.perf_counter())
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.total_time = time.perf_counter() - self.times[0]
        print(self._make_str("Total", self.total_time))

    def time(self, name: Optional[str] = None):
        self.times.append(time.perf_counter())
        name = name or ""
        print(self._make_str(name, time.perf_counter() - self.times[-2]))

    def __call__(self, f: Callable):
        @wraps(f)
        def wrapper(*args, **kwargs):
            self.times.append(time.perf_counter())
            result = f(*args, **kwargs)
            print(self._make_str(f.__name__, time.perf_counter() - self.times[-1]))
            return result

        return wrapper

    @staticmethod
    def _make_str(name: str, t: float):
        return f"[Timer] {name}: {t} sec"




from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field
from enum import Enum
from math import ceil
from typing import List, Optional, Callable, TypeVar

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from numpy import ndarray


class PlotType(Enum):
    IMAGE = "image"
    PLOT = "plot"


class SubPlot(ABC):
    plot_type: PlotType


@dataclass
class ImagePlot:
    image: ndarray
    title: Optional[str] = None
    kwargs: dict = field(default_factory=dict)
    plot_type = PlotType.IMAGE


@dataclass
class PlotPlot:
    plot_func: Callable[[Axes], Axes]
    subplot_kwargs: dict = field(default_factory=dict)
    plot_type = PlotType.PLOT


P = TypeVar("P", bound=SubPlot)


class MultiPlots:
    def __init__(self, n_row: Optional[int] = None, n_col: Optional[int] = None,
        figure: Optional[Figure] = None,
        title: Optional[str] = None,
        add_by_row: bool = True,
    ):
        self.n_row = n_row
        self.n_col = n_col
        self.figure = figure or plt.figure()
        self.title = title
        self.subplots: List[P] = []
        self.add_by_row = add_by_row

    def add_image(self, image: ndarray, title: Optional[str] = None, **kwargs) -> MultiPlots:
        self.subplots.append(ImagePlot(image, title, kwargs))
        return self

    def add_plot(self, plot_func: Callable[[Axes], Axes], **kwargs) -> MultiPlots:
        self.subplots.append(PlotPlot(plot_func, kwargs))
        return self

    def _determine_shape(self) -> None:
        if self.n_row is None and self.n_col is None:
            if self.add_by_row:
                self.n_row = 1
                self.n_col = len(self.subplots)
            else:
                self.n_row = len(self.subplots)
                self.n_col = 1
        elif self.n_row is None and self.n_col is not None:
            self.n_row = ceil(len(self.subplots) / self.n_col)
        elif self.n_col is None and self.n_row is not None:
            self.n_col = ceil(len(self.subplots) / self.n_row)
        else:
            if self.n_row * self.n_col < len(self.subplots):
                raise ValueError("n_row * n_col must be greater than the number of subplots")

    def _location(self, i: int) -> int:
        if self.add_by_row:
            return i + 1
        else:
            return i // self.n_col + 1 + (i % self.n_col) * self.n_row

    def plot(self) -> Figure:
        self._determine_shape()
        for i, subplot in enumerate(self.subplots):
            if subplot.plot_type == PlotType.IMAGE:
                ax = self.figure.add_subplot(self.n_row, self.n_col, self._location(i))
                ax.imshow(subplot.image, **subplot.kwargs)
                ax.set_axis_off()
                if subplot.title is not None:
                    ax.set_title(subplot.title)
            elif subplot.plot_type == PlotType.PLOT:
                ax = self.figure.add_subplot(self.n_row, self.n_col, self._location(i), **subplot.subplot_kwargs)
                ax = subplot.plot_func(ax)
        if self.title is not None:
            self.figure.suptitle(self.title)
        return self.figure



class TextWriter:
    """TextWriter for writing files for text"""

    @staticmethod
    def to_plain(path: str, text: str, encoding: str = "UTF-8") -> None:
        """
        Save as a plain text file.

        Args:
            path (``str``): The path of output file.
            text (``str``): The content for output.
            encoding (``str``, optional): Python file IO encoding parameter. Default: "UTF-8".
        """
        with open(path, "w", encoding=encoding) as file:
            file.write(text)

    to = to_plain

    def __new__(cls, path: str, text: str, encoding: str = "UTF-8") -> None:
        """Alias of :meth:`~TextWriter.to`"""
        cls.to(path, text, encoding)



from typing import Union

import numpy as np
from numpy import ndarray

from tensorneko_util.util import dispatch


@dispatch
def crop_with_padding(image: ndarray, x1: int, x2: int, y1: int, y2: int, pad_value: Union[int, float] = 0.,
    batch: bool = False
) -> ndarray:
    """
    Crop image with padding.

    This function is implemented for :class:`~numpy.ndarray`.

    Args:
        image (:class:`~numpy.ndarray`): Image to be cropped. The shape should be ([B,] H, W) or ([B,] H, W, C).
        x1 (``int``): Left-top x-coordinate of the cropped area.
        x2 (``int``): Right-bottom x-coordinate of the cropped area.
        y1 (``int``): Left-top y-coordinate of the cropped area.
        y2 (``int``): Right-bottom y-coordinate of the cropped area.
        pad_value (``int`` | ``float``, optional): The value to fill the padded region. Default 0.
        batch (``bool``, optional): Indicate if the input is a batch of images. Default False.

    Returns:
        :class:`~numpy.ndarray`: Cropped image.

    """
    assert y2 > y1 and x2 > x1, "Should follow y2 > y1 and x2 > x1"

    if not batch:
        image = image[np.newaxis, ...]

    crop_shape = np.array([y2 - y1, x2 - x1])

    if len(image.shape) == 3:
        b, h, w = image.shape
        cropped = np.full((b, *crop_shape), pad_value, dtype=image.dtype)
    elif len(image.shape) == 4:
        b, h, w, c = image.shape
        cropped = np.full((b, *crop_shape, c), pad_value, dtype=image.dtype)
    else:
        raise ValueError("Invalid shape, the image should be one of following shapes: ([B,] H, W) or ([B,] H, W, C)")

    # compute cropped index of image
    image_y_start, image_x_start = np.clip([y1, x1], 0, [h, w])
    image_y_end, image_x_end = np.clip([y2, x2], 0, [h, w])

    # compute target index of output
    crop_y_start, crop_x_start = np.clip([-y1, -x1], 0, crop_shape)
    crop_y_end, crop_x_end = crop_shape - np.clip([y2 - h, x2 - w], 0, crop_shape)

    # assign values
    cropped[:, crop_y_start:crop_y_end, crop_x_start:crop_x_end] = \
        image[:, image_y_start:image_y_end, image_x_start:image_x_end]

    return cropped if batch else cropped[0]


@dispatch
def crop_with_padding(image: ndarray, x1: np.int32, x2: np.int32, y1: np.int32, y2: np.int32,
    pad_value: Union[int, float] = 0., batch: bool = False
) -> ndarray:
    return crop_with_padding(image, int(x1), int(x2), int(y1), int(y2), pad_value, batch)



from .bus import EventBus


def subscribe(func):
    return EventBus.default.subscribe(func)


def subscribe_async(func):
    return EventBus.default.subscribe_async(func)


def subscribe_thread(func):
    return EventBus.default.subscribe_thread(func)


def subscribe_process(func):
    return EventBus.default.subscribe_process(func)



import pathlib
from typing import Optional, Union

import numpy as np
from einops import rearrange

from .video_data import VideoData
from .._default_backends import _default_video_io_backend
from ...backend.visual_lib import VisualLib
from ...util import dispatch
from ...util.type import T_ARRAY


class VideoWriter:
    """The VideoWriter for writing video file"""

    @classmethod
    @dispatch
    def to(cls, path: str, video: VideoData, audio_codec: str = None, channel_first: bool = False,
        backend: VisualLib = None
    ) -> None:
        """
        Write to video file from :class:`~tensorneko.io.video.video_data.VideoData`.

        Args:
            path (``str``): The path of output file.
            video (:class:`~tensorneko.io.video.video_data.VideoData`): The VideoData object for output.
            audio_codec (``str``, optional): The audio codec if audio is required. Default: None.
            channel_first (``bool``, optional): Get image dimension (T, H, W, C) if False or (T, C, H, W) if True.
                Default: False.
            backend (:class:`~tensorneko.backend.visual_lib.VisualLib`, optional): VisualLib backend.
                Default: opencv if installed, then torchvision if installed, then ffmpeg if available.
        """
        cls.to(path, video.video, video.info.video_fps, video.audio, video.info.audio_fps, audio_codec, channel_first, backend)

    @classmethod
    @dispatch
    def to(cls, path: str, video: T_ARRAY, video_fps: float, audio: T_ARRAY = None,
        audio_fps: Optional[int] = None, audio_codec: str = None, channel_first: bool = False,
        backend: VisualLib = None
    ) -> None:
        """
        Write to video file from :class:`~torch.Tensor` or :class:`~numpy.ndarray` with (T, C, H, W).

        Args:
            path (``str``): The path of output file.
            video (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): The video tensor or array with (T, C, H, W) for
                output.
            video_fps (``float``): The video fps.
            audio (:class:`~torch.Tensor` | :class:`~numpy.ndarray`, optional): The audio tensor or array with (C, T)
                for output. None means no audio in output video file. Default: None.
            audio_fps (``int``, optional): The audio fps. Default: None.
            audio_codec (``str``, optional): The audio codec if audio is required. Default: None.
            channel_first (``bool``, optional): Get image dimension (T, H, W, C) if False or (T, C, H, W) if True.
                Default: False.
            backend (:class:`~tensorneko.backend.visual_lib.VisualLib`, optional): VisualLib backend.
                Default: opencv if installed, then torchvision if installed, then ffmpeg if available.
        """
        if audio:
            if audio_codec is None:
                raise ValueError("audio_codec is required if audio is required.")
        backend = backend or _default_video_io_backend()
        if channel_first:
            video = rearrange(video, "t c h w -> t h w c")

        # to uint8
        if isinstance(video, np.ndarray):
            video = (video * 255).astype(np.uint8)
        else:
            try:
                import torch
                if isinstance(video, torch.Tensor):
                    if backend == VisualLib.PYTORCH:
                        video = (video * 255).type(torch.IntTensor)
                    else:
                        video = (video.numpy() * 255).astype(np.uint8)
                else:
                    raise ValueError("Unknown data type. The image array type must be numpy.ndarray or torch.Tensor.")
            except ImportError:
                raise ValueError("Unknown data type. The image array type must be numpy.ndarray or torch.Tensor.")

        ext = pathlib.Path(path).suffix

        if backend == VisualLib.OPENCV:
            if audio is not None:
                raise ValueError("Write audio is not supported in opencv backend.")
            if not VisualLib.opencv_available():
                raise ValueError("Opencv is not installed.")
            import cv2
            if ext == ".avi":
                fourcc = cv2.VideoWriter_fourcc(*"XVID")
            elif ext == ".mp4":
                fourcc = cv2.VideoWriter_fourcc(*"mp4v")
            else:
                raise NotImplementedError("Only .avi and .mp4 are supported in opencv backend.")
            writer = cv2.VideoWriter(path, fourcc, video_fps, (video.shape[2], video.shape[1]))
            for i in range(video.shape[0]):
                writer.write(cv2.cvtColor(video[i], cv2.COLOR_RGB2BGR))
            writer.release()
        elif backend == VisualLib.PYTORCH:
            if not VisualLib.pytorch_available():
                raise ValueError("Torchvision is not installed.")
            import torchvision
            torchvision.io.write_video(path, rearrange(video), video_fps, audio_fps=audio_fps, audio_array=audio)
        elif backend == VisualLib.FFMPEG:
            if audio is not None:
                raise ValueError("Write audio is not supported in ffmpeg backend.")
            if not VisualLib.ffmpeg_available():
                raise ValueError("FFMPEG is not installed.")
            import ffmpeg
            pipe = ffmpeg.input("pipe:", format="rawvideo", pix_fmt="rgb24", s=f"{video.shape[2]}x{video.shape[1]}") \
                .output(path, format=ext[1:], pix_fmt="rgb24", r=video_fps) \
                .overwrite_output() \
                .run_async(pipe_stdin=True)

            for i in range(video.shape[0]):
                pipe.stdin.write(video[i].tobytes())

            pipe.stdin.close()
            pipe.wait()
        else:
            raise ValueError("Unknown backend. Should be one of VisualLib.OPENCV, VisualLib.PYTORCH, VisualLib.FFMPEG.")

    def __new__(cls, path: str, video: Union[T_ARRAY, VideoData], *args, **kwargs):
        """Alias of :func:`~tensorneko.io.video.video_io.to`."""
        return cls.to(path, video, *args, **kwargs)



from __future__ import annotations

from typing import Generic, Dict, Optional, Union, Tuple

from ..util.type import K, V


class BiMap(Generic[K, V]):
    def __init__(self, items: Dict[K, V] = None):
        self.forward: Dict[K, V] = {}
        self.backward: Dict[V, K] = {}
        if items is not None:
            for key, value in items.items():
                self.forward[key] = value
                self.backward[value] = key

    def __setitem__(self, key: K, value: V):
        self.forward[key] = value
        self.backward[value] = key

    def __getitem__(self, key: K) -> V:
        return self.forward[key]

    def get(self, key: K, default: Optional[V] = None) -> V:
        if default is None:
            return self.forward[key]
        else:
            return self.forward.get(key, default)

    def get_key(self, value: V, default: Optional[K] = None) -> K:
        if default is None:
            return self.backward[value]
        else:
            return self.backward.get(value, default)

    def __len__(self):
        return len(self.forward)

    def __contains__(self, key: K) -> bool:
        return key in self.forward

    def __delitem__(self, key: K):
        del self.backward[self.forward[key]]
        del self.forward[key]

    def clear(self):
        self.forward.clear()
        self.backward.clear()

    def __copy__(self):
        bimap = BiMap()
        bimap.forward = self.forward.copy()
        bimap.backward = self.backward.copy()

    def items(self):
        return self.forward.items()

    def keys(self):
        return self.forward.keys()

    def values(self):
        return self.forward.values()

    def __iter__(self):
        return self.forward.__iter__()

    def update(self, other: Union[Dict[K, V], BiMap[K, V]]):
        if isinstance(other, BiMap):
            self.forward.update(other.forward)
            self.backward.update(other.backward)
        else:
            self.forward.update(other)
            self.backward.update({v: k for k, v in other.items()})

    def pop(self, key: K) -> V:
        value = self.forward.pop(key)
        self.backward.pop(value)
        return value

    def popitem(self) -> Tuple[K, V]:
        key, value = self.forward.popitem()
        self.backward.pop(value)
        return key, value

    def __repr__(self):
        return f"BiMap({self.forward})"

    def __str__(self):
        return f"BiMap({self.forward})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BiMap):
            return self.forward == other.forward and self.backward == other.backward
        else:
            return False

    def __hash__(self):
        return hash(self.forward) ^ hash(self.backward)

    def __bool__(self):
        return bool(self.forward)



from itertools import islice, takewhile
from typing import Optional

import numpy as np
from einops import rearrange

from .video_data import VideoData
from .._default_backends import _default_video_io_backend
from ...backend.visual_lib import VisualLib
from ...util import dispatch


class VideoReader:

    @classmethod
    def of(cls, path: str, channel_first: bool = True, backend: Optional[VisualLib] = None) -> VideoData:
        """
        Read video array from given file.

        Args:
            path (``str``): Path to the video file.
            channel_first (``bool``, optional): Get image dimension (T, H, W, C) if False or (T, C, H, W) if True.
                Default: True.
            backend (:class:`~tensorneko.backend.visual_lib.VisualLib`, optional): VisualLib backend.
                Default: opencv if installed, then torchvision if installed, then ffmpeg if available.

        Returns:
            :class:`~tensorneko.io.video.video_data.VideoData`:
                A VideoData object contains a float tensor of video, with value range of 0. to 1,
                an audio tensor of (T, C) and a :class:`~tensorneko.io.video.video_data.VideoInfo` contains fps info.
        """
        backend = backend or _default_video_io_backend()
        if backend == VisualLib.OPENCV:
            if not VisualLib.opencv_available():
                raise ValueError("OpenCV is not installed.")
            import cv2
            cap = cv2.VideoCapture(path)
            if not cap.isOpened():
                raise ValueError("Failed to open video file: {}".format(path))
            fps = cap.get(cv2.CAP_PROP_FPS)
            frames = []
            while True:
                ret, frame = cap.read()
                if not ret:
                    break
                else:
                    frames.append(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
            cap.release()
            video = np.stack(frames, axis=0)
            if channel_first:
                video = rearrange(video, 'T H W C -> T C H W')
            return VideoData(video, None, {"video_fps": fps})
        elif backend == VisualLib.PYTORCH:
            if not VisualLib.pytorch_available():
                raise ValueError("Torchvision is not installed.")
            import torchvision
            video, audio, info = torchvision.io.read_video(path)
            if channel_first:
                video = rearrange(video, 'T H W C -> T C H W')
            audio = rearrange(audio, 'C T -> T C')
            return VideoData(video, audio, info)
        elif backend == VisualLib.FFMPEG:
            if not VisualLib.ffmpeg_available():
                raise ValueError("FFMPEG is not installed.")
            import ffmpeg

            probe = ffmpeg.probe(path)
            for each in probe['streams']:
                if each['codec_type'] == 'video':
                    fps = each['avg_frame_rate']
                    height = each['height']
                    width = each['width']
                    break
            else:
                raise ValueError("Failed to get video from {}".format(path))

            out, _ = ffmpeg.input(path).output('pipe:', format='rawvideo', pix_fmt='rgb24').run(quiet=True)
            video = np.frombuffer(out, np.uint8).reshape([-1, height, width, 3])
            if channel_first:
                video = rearrange(video, 'T H W C -> T C H W')
            return VideoData(video, None, {"video_fps": fps})
        else:
            raise ValueError("Unknown backend: {}".format(backend))

    @classmethod
    def with_indexes(cls, path: str, indexes: np.ndarray,
        channel_first: bool = True, backend: Optional[VisualLib] = None
    ) -> VideoData:
        """
        Get a video frames with indexes. The audio will be ignored.
        Args:
            path (``str``): Path to the video file.
            indexes (``np.ndarray``): Indexes of the video.
            channel_first (``bool``, optional): Get image dimension (T, H, W, C) if False or (T, C, H, W) if True.
                Default: True.
            backend (:class:`~tensorneko.backend.visual_lib.VisualLib`, optional): VisualLib backend.
                Default: opencv if installed, then torchvision if installed, then ffmpeg if available.
                Note: The ffmpeg backend cannot precisely index the frame. Recommend to use pytorch backend.
        Returns:
            :class:`~tensorneko.io.video.video_data.VideoData`:
                A VideoData object contains a float tensor of video, with value range of 0. to 1,
                an audio tensor of (T, C) and a :class:`~tensorneko.io.video.video_data.VideoInfo` contains fps info.
        """
        backend = backend or _default_video_io_backend()
        if backend == VisualLib.OPENCV:
            if not VisualLib.opencv_available():
                raise ValueError("OpenCV is not installed.")
            import cv2
            cap = cv2.VideoCapture(path)
            if not cap.isOpened():
                raise ValueError("Failed to open video file: {}".format(path))
            fps = cap.get(cv2.CAP_PROP_FPS)
            frames = []
            i = 0
            while True:
                ret, frame = cap.read()
                if not ret:
                    break
                else:
                    if i in indexes:
                        frames.append(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
                if i == indexes[-1]:
                    # early stop
                    break
            assert len(frames) == len(indexes)
            cap.release()
            video = np.stack(frames, axis=0)
            if channel_first:
                video = rearrange(video, 'T H W C -> T C H W')
            return VideoData(video, None, {"video_fps": fps})
        elif backend == VisualLib.PYTORCH:
            if not VisualLib.pytorch_available():
                raise ValueError("Torchvision is not installed.")
            import torchvision
            reader = torchvision.io.VideoReader(path)
            meta = reader.get_metadata()["video"]
            fps = meta["fps"]
            first_index = indexes[0]
            first_pts = first_index / fps
            last_pts = indexes[-1] / fps
            indexes = indexes - first_index
            video = []
            for frame in takewhile(lambda x: x["pts"] <= last_pts, reader.seek(first_pts)):
                video.append(frame["data"])
            video = np.stack(video)
            video = video[indexes]  # (T C H W)

            if not channel_first:
                video = rearrange(video, 'T C H W -> T H W C')
            return VideoData(video, None, {"video_fps": fps})
        elif backend == VisualLib.FFMPEG:
            if not VisualLib.ffmpeg_available():
                raise ValueError("FFMPEG is not installed.")
            import ffmpeg

            probe = ffmpeg.probe(path)
            for each in probe['streams']:
                if each['codec_type'] == 'video':
                    fps = each['avg_frame_rate']
                    height = each['height']
                    width = each['width']
                    break
            else:
                raise ValueError("Failed to get video from {}".format(path))

            out = ffmpeg.input(path) \
                .filter('select', "+".join(f"eq(n,{f})" for f in indexes)) \
                .output('pipe:', vframes=len(indexes), format='rawvideo', pix_fmt='rgb24') \
                .run(quiet=True)[0]
            video = np.frombuffer(out, np.uint8).reshape([-1, height, width, 3])
            if channel_first:
                video = rearrange(video, 'T H W C -> T C H W')
            return VideoData(video, None, {"video_fps": fps})
        else:
            raise ValueError("Unknown backend: {}".format(backend))

    @classmethod
    @dispatch
    def with_range(cls, path: str, start: int, end: int, step: int, channel_first: bool = True,
        backend: Optional[VisualLib] = None
    ) -> VideoData:
        backend = backend or _default_video_io_backend()
        if backend == VisualLib.PYTORCH:
            if not VisualLib.pytorch_available():
                raise ValueError("Torchvision is not installed.")
            import torchvision
            reader = torchvision.io.VideoReader(path)
            meta = reader.get_metadata()["video"]
            fps = meta["fps"]
            first_pts = start / fps

            video = []
            for frame in islice(reader.seek(first_pts), start, end, step):
                video.append(frame["data"])
            video = np.stack(video)  # (T C H W)

            if not channel_first:
                video = rearrange(video, 'T C H W -> T H W C')
            return VideoData(video, None, {"video_fps": fps})
        else:
            cls.with_indexes(path, np.arange(start, end, step), channel_first, backend)

    @classmethod
    @dispatch
    def with_range(cls, path: str, end: int, channel_first: bool = True,
        backend: Optional[VisualLib] = None
    ) -> VideoData:
        return cls.with_range(path, 0, end, 1, channel_first, backend)

    def __new__(cls, path: str, channel_first: bool = False, backend: Optional[VisualLib] = None) -> VideoData:
        """Alias of :meth:`~.VideoReader.of`"""
        return cls.of(path, channel_first, backend)



from .visual_lib import VisualLib
from .audio_lib import AudioLib
from .blocking import run_blocking
from . import parallel

__all__ = [
    "VisualLib",
    "AudioLib",
    "run_blocking",
    "parallel",
]



_is_tqdm_available = None


def import_tqdm():
    global _is_tqdm_available
    if _is_tqdm_available is None:
        try:
            import tqdm
            _is_tqdm_available = True
            return tqdm
        except ImportError:
            _is_tqdm_available = False
            raise ImportError("tqdm is not installed. Please install it by `pip install tqdm`")
    else:
        if _is_tqdm_available:
            import tqdm
            return tqdm
        else:
            raise ImportError("tqdm is not installed. Please install it by `pip install tqdm`")


def import_tqdm_auto():
    global _is_tqdm_available
    if _is_tqdm_available is None:
        try:
            from tqdm import auto
            _is_tqdm_available = True
            return auto
        except ImportError:
            _is_tqdm_available = False
            raise ImportError("tqdm is not installed. Please install it by `pip install tqdm`")
    else:
        if _is_tqdm_available:
            from tqdm import auto
            return auto
        else:
            raise ImportError("tqdm is not installed. Please install it by `pip install tqdm`")



import json
from typing import Union, Optional

from ...util.type import T


class JsonReader:

    @classmethod
    def of(cls, path: str, clazz: Optional[T] = None, encoding: str = "UTF-8") -> Union[T, dict, list]:
        """
        Read json files to ``list`` or ``dict``.

        Args:
            path (``str``): Json file path.
            clazz: (``T``, optional): The object of the json read for. The type should be decorated by
                :func:`json_data`. This should be ``T`` or ``List[T]`` or ``List[List[T]]``
            encoding (``str``, optional): The encoding for python ``open`` function. Default: "UTF-8"

        Returns:
            ``dict`` | ``list`` | ``object``: The object of given json.
        """
        if clazz is not None:
            if "typing.List[" in str(clazz):
                with open(path, "r", encoding=encoding) as file:
                    obj: list = json.loads(file.read())

                inner_type = clazz.__args__[0]
                if "typing.List[" in str(inner_type):
                    inner_inner_type = inner_type.__args__[0]
                    try:
                        is_json_data = inner_type.is_json_data
                    except AttributeError:
                        is_json_data = False
                    if is_json_data:
                        obj = [[inner_inner_type(e) for e in each] for each in obj]

                else:
                    try:
                        is_json_data = inner_type.is_json_data
                    except AttributeError:
                        is_json_data = False

                    if is_json_data:
                        obj = list(map(inner_type, obj))
            else:
                with open(path, "r", encoding=encoding) as file:
                    obj = clazz(json.loads(file.read()))
        else:
            with open(path, "r", encoding=encoding) as file:
                obj = json.loads(file.read())

        return obj

    def __new__(cls, path: str, clazz: T = None, encoding: str = "UTF-8") -> Union[T, dict, list]:
        """Alias of :meth:`~tensorneko_util.io.json.json_reader.JsonReader.of`."""
        return cls.of(path, clazz, encoding)



import os.path
from typing import Optional, Union, Tuple

from numpy import ndarray

from .abstract_face_detector import AbstractFaceDetector
from ..crop import crop_with_padding
from ...io import read


class FaceXZooFaceDetector(AbstractFaceDetector):
    """
    The face sdk is in https://github.com/ControlNet/FaceX-Zoo.
    The fork of https://github.com/JDAI-CV/FaceX-Zoo/tree/main/face_sdk.

    Args:
        face_sdk_path (``str``): The path to the face_sdk path of FaceXZoo.
        device (``str``, optional): The device to use, default is "cuda:0".

    Examples::

        from tensorneko_util.preprocess.face_detector import FaceXZooFaceDetector

        detector = FaceXZooFaceDetector(FaceXZooFaceDetector.install(), "cuda:0")

        # process image, save to facexzoo_face_detector_out.jpg
        detector.process_image("test/resource/image_sample/1.183.jpg",
            "test/resource/temp_output/facexzoo_face_detector_out.jpg")

        # process video, save to facexzoo_face_detector_out.mp4
        detector.process_video("test/resource/video_sample/sample.mp4",
            "test/resource/temp_output/facexzoo_face_detector_out.mp4")

    """

    def __init__(self, face_sdk_path: Optional[str] = None, device: str = "cuda:0"):
        try:
            import cv2
        except ImportError:
            raise ImportError("OpenCV and opencv-contrib-python is not installed. Please install it first.")

        self.cv2 = cv2

        if face_sdk_path is not None:
            import sys
            sys.path.append(face_sdk_path)
        try:
            from core.model_handler.face_detection.FaceDetModelHandler import FaceDetModelHandler
            from core.model_loader.face_detection.FaceDetModelLoader import FaceDetModelLoader
        except ImportError:
            raise ImportError("FaceX-Zoo cannot be imported, please specify the path to the face_sdk path of FaceXZoo"
                              " or put it in the working directory.")

        model_conf = read.yaml(os.path.join(face_sdk_path, "config", "model_conf.yaml"))
        model_path = os.path.join(face_sdk_path, 'models')
        scene = 'non-mask'
        model_category = 'face_detection'
        model_name = model_conf[scene][model_category]

        faceDetModelLoader = FaceDetModelLoader(model_path, model_category, model_name)
        model, cfg = faceDetModelLoader.load_model()
        self.faceDetModelHandler = FaceDetModelHandler(model, device, cfg)

    @staticmethod
    def install(path: Optional[str] = None) -> str:
        """
        Install FaceX-Zoo by clone from GitHub.

        Args:
            path (``str``, optional): The path to install FaceX-Zoo, default is "./FaceX-Zoo".

        Returns:
            ``str``: The path to the installed FaceX-Zoo.

        """
        path = path or "FaceX-Zoo"
        if os.path.exists(path):
            return path

        os.system(f"git clone --depth=1 https://github.com/ControlNet/FaceX-Zoo {path or ''}")
        return path

    def detect_face(self, image: ndarray, *args, **kwargs):
        assert image.ndim == 3 and image.shape[2] == 3, "frame should be 3-dim"
        dets = self.faceDetModelHandler.inference_on_image(image)
        return dets

    def process_image(self, image_path: str, out_path: str, max_faces=1, margin=0, *args, **kwargs):
        if max_faces > 1:
            raise NotImplementedError("Multiple faces are not supported yet.")

        frame = self.cv2.imread(image_path)
        dets = self.detect_face(frame)
        for det in dets[:max_faces]:
            x1, y1, x2, y2, confidence = det
            cropped = crop_with_padding(frame, int(x1 - margin), int(x2 + margin), int(y1 - margin), int(y2 + margin))
            self.cv2.imwrite(out_path, cropped)

    def _infer_frame_size(self, video_path: str, margin: int = 0) -> Tuple[int, int]:
        video = self.cv2.VideoCapture(video_path)
        while True:
            ret, frame = video.read()
            if not ret:
                break
            dets = self.detect_face(frame)
            if len(dets) > 0:
                x1, y1, x2, y2, confidence = dets[0]
                # center
                side = int(max(abs(x2 - x1), abs(y2 - y1)))
                video.release()
                return side + 2 * margin, side + 2 * margin

    def process_video(self, video_path: str, out_path: str, frame_size: Optional[Union[int, Tuple[int, int]]] = None,
        margin=0, fourcc="mp4v", *args, **kwargs
    ):
        video = self.cv2.VideoCapture(video_path)
        if not video.isOpened():
            raise IOError("Cannot open video file: " + video_path)

        # infer frame size
        if frame_size is None:
            frame_size = self._infer_frame_size(video_path, margin)

        if type(frame_size) is int:
            frame_size = frame_size, frame_size

        fps = video.get(self.cv2.CAP_PROP_FPS)
        writer = self.cv2.VideoWriter(out_path, self.cv2.VideoWriter_fourcc(*fourcc), fps, frame_size)
        x1, y1, x2, y2 = 0, 0, 1, 1
        while True:
            ret, frame = video.read()
            if not ret:
                break
            dets = self.detect_face(frame)
            if len(dets) > 0:
                x1, y1, x2, y2, confidence = dets[0]
                # center
                x, y = int((x1 + x2) / 2), int((y1 + y2) / 2)
                side = int(max(abs(x2 - x1), abs(y2 - y1)))
                x1 = x - side // 2
                x2 = x + side // 2
                y1 = y - side // 2
                y2 = y + side // 2

            cropped = crop_with_padding(frame, int(x1 - margin), int(x2 + margin), int(y1 - margin), int(y2 + margin))
            resized = self.cv2.resize(cropped, frame_size)
            writer.write(resized)
        video.release()
        writer.release()



from contextlib import contextmanager
from typing import TypeVar

from .bus import EventBus


class EventMeta(type):

    def __call__(cls, *args, bus=EventBus.default, **kwargs):
        event = super().__call__(*args, **kwargs)
        bus.emit(event, blocking=_blocking_flag)
        return event


_blocking_flag = True


@contextmanager
def no_blocking():
    """
    Context manager to disable blocking event handler.
    """
    global _blocking_flag
    _blocking_flag = False
    yield
    _blocking_flag = True


class Event(metaclass=EventMeta):
    bus: EventBus


E = TypeVar("E", bound=Event)



from typing import Union

from .audio_data import AudioData
from .._default_backends import _default_audio_io_backend
from ...backend.audio_lib import AudioLib
from ...util import dispatch
from ...util.type import T_ARRAY


class AudioWriter:
    """AudioWriter for writing audio file"""

    @classmethod
    @dispatch
    def to(cls, path: str, audio: AudioData, channel_first: bool = True, backend: AudioLib = None) -> None:
        """
        Save wav file from :class:`~tensorneko.io.audio.audio_data.AudioData`.

        Args:
            path (``str``): The path of output file.
            audio (:class:`~tensorneko.io.audio.AudioData`): The AudioData object for output.
            channel_first (``bool``, optional): Whether the audio is channel first. The input shape is (C, T) if true
                and (T, C) if false. Default: True.
            backend (:class:`~tensorneko.io.audio.audio_lib.AudioLib`, optional): The audio library to use.
                Default: pytorch.
        """
        return cls.to(path, audio.audio, audio.sample_rate, channel_first, backend)

    @classmethod
    @dispatch
    def to(cls, path: str, audio: T_ARRAY, sample_rate: int = 16000, channel_first: bool = True,
        backend: AudioLib = None
    ):
        """
        Save wav file from :class:`~torch.Tensor` or :class:`~numpy.ndarray` with (C, T).

        Args:
            path (``str``): The path of output file.
            audio (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): The tensor or array of audio.
            sample_rate (``int``, optional): The sample rate of the audio. Default: 16000.
            channel_first (``bool``, optional): Whether the audio is channel first. The input shape is (C, T) if true
                and (T, C) if false. Default: True.
            backend (:class:`~tensorneko.io.audio.audio_lib.AudioLib`, optional): The audio library to use.
                Default: pytorch.
        """
        backend = backend or _default_audio_io_backend()

        if backend == AudioLib.PYTORCH:
            if not AudioLib.pytorch_available():
                raise ValueError("Torchaudio is not available.")
            import torchaudio
            torchaudio.save(path, audio, sample_rate, channels_first=channel_first)
        else:
            raise ValueError("Unknown audio library: {}".format(backend))

    def __new__(cls, path: str, audio: Union[AudioData, T_ARRAY], *args, **kwargs):
        """Alias to :meth:`~AudioWriter.to`"""
        return cls.to(path, audio, *args, **kwargs)



from enum import Enum
from typing import Optional


class AudioLib(Enum):
    PYTORCH = 1

    _is_torchaudio_available: Optional[bool] = None

    @classmethod
    def pytorch_available(cls) -> bool:
        if cls._is_torchaudio_available is None:
            try:
                import torchaudio
                cls._is_torchaudio_available = True
            except ImportError:
                cls._is_torchaudio_available = False
        return cls._is_torchaudio_available



from typing import Type, TypeVar, Dict, Any

import yaml
from yaml import Loader as Loader_, BaseLoader, FullLoader, SafeLoader, UnsafeLoader

L = TypeVar("L", Loader_, BaseLoader, FullLoader, SafeLoader, UnsafeLoader)


class YamlReader:

    @classmethod
    def of(cls, path: str, Loader: Type[L] = Loader_) -> Dict[str, Any]:
        """
        Load a YAML file and return a Python dict.
        Args:
            path (``str``): YAML file path.
            Loader: (``L``, optional): The loader for yaml. Default: :class:`yaml.Loader`

        Returns:
            ``dict``: The Python dict of given YAML file.
        """
        with open(path, "r") as file:
            return yaml.load(file, Loader=Loader)

    def __new__(cls, path: str, Loader: Type[L] = Loader_):
        """Alias of :meth:`~tensorneko_util.io.yaml.yaml_reader.YamlReader.of`."""
        return cls.of(path, Loader)



from __future__ import annotations

from abc import ABC
from typing import Generic, Callable, TYPE_CHECKING, Optional

from tensorneko_util.util.dispatcher import dispatch
from .type import P

if TYPE_CHECKING:
    from ..visualization.watcher import Component


class Ref(ABC, Generic[P]):
    _value: P

    def __init__(self, value: P):
        self._value = value
        self.bound_comp: Optional[Component] = None

    @property
    def value(self) -> P:
        return self._value

    @value.setter
    def value(self, value: P):
        self._value = value
        if self.bound_comp is not None:
            self.bound_comp.update_view()

    def apply(self, f: Callable[[P], P]) -> Ref[P]:
        new_value = f(self.value)
        assert type(new_value) is type(self.value), "The transform should be the same type."
        self.value = new_value
        return self

    def __rshift__(self, f: Callable[[P], P]) -> Ref[P]:
        return self.apply(f)


class StringRef(Ref[str]):

    def __str__(self):
        return self.value


class IntRef(Ref[int]):

    def __int__(self):
        return self.value


class FloatRef(Ref[float]):

    def __float__(self):
        return self.value


class BoolRef(Ref[bool]):

    def __bool__(self):
        return self.value


@dispatch.of(str)
def ref(value: str) -> StringRef:
    return StringRef(value)


@dispatch.of(int)
def ref(value: int) -> IntRef:
    return IntRef(value)


@dispatch.of(float)
def ref(value: float) -> FloatRef:
    return FloatRef(value)


@dispatch.of(bool)
def ref(value: bool) -> BoolRef:
    return BoolRef(value)



from __future__ import annotations

from functools import reduce
from itertools import chain
from sys import maxsize
from typing import List, Callable, Optional, Union, Iterable, Generator

from .abstract_seq import AbstractSeq
from ...type import T, R
from ....backend.parallel import ParallelType, ExecutorPool


class Stream(AbstractSeq[T]):

    def __init__(self, iterable: Iterable[T] = iter(())):
        self._iter = iter(iterable)
        self._cache: List[T] = []
        self._finished = False
        if isinstance(iterable, Stream):
            self._cache = iterable._cache

    @staticmethod
    def of(*items: T) -> Stream[T]:
        return Stream(items)

    @staticmethod
    def from_stream(stream: Stream[T]) -> Stream[T]:
        stream = Stream(stream._iter)
        stream._cache = stream._cache
        return stream

    def __getitem__(self, item: Union[int, slice]) -> Union[T, AbstractSeq[T]]:
        if isinstance(item, int):
            if item < 0:
                raise ValueError("Negative slice is not allowed in Stream")
            for _ in self._iter_to(item):
                pass

        elif isinstance(item, slice):
            low, high, step = item.indices(maxsize)
            if step == 0:
                raise ValueError("Step must not be 0")
            return self.__class__() << map(self.__getitem__, range(low, high, step or 1))
        else:
            raise TypeError("Invalid argument type")

        return self._cache.__getitem__(item)

    def __lshift__(self, right_iter: Iterable[T]) -> Stream[T]:
        self._iter = chain(self._iter, right_iter)
        return self

    def __iter__(self) -> Generator[T, None, None]:
        return self._iter_all()

    def _iter_to(self, index: int, start: int = 0) -> Generator[T, None, None]:
        return self._iter_cond(lambda i: i <= index, start)

    def _iter_all(self) -> Generator[T, None, None]:
        return self._iter_cond(lambda i: True)

    def _iter_cond(self, cond: Callable[[int], bool], start: int = 0) -> Generator[T, None, None]:
        i = start
        while cond(i):
            if self._cache_size > i:
                yield self._cache[i]
            elif self._cache_size == i:
                try:
                    item = next(self._iter)
                except StopIteration:
                    self._finished = True
                    break
                self._cache.append(item)
                yield item
            else:
                for item in self._iter_to(i, start=self._cache_size):
                    yield item
            i += 1

    @property
    def _cache_size(self) -> int:
        return len(self._cache)

    def map(self, f: Callable[[T], R]) -> Stream[R]:
        return MapStream(self, f)

    def for_each(self, f: Callable[[T], None], progress_bar: bool = False, parallel_type: Optional[ParallelType] = None,
        **tqdm_args
    ) -> None:
        if parallel_type is None:
            items = self if not progress_bar else self._tqdm(self, **tqdm_args)
            for item in items:
                f(item)
        else:
            futures = []

            if f.__name__ == "<lambda>":
                raise NotImplementedError("lambda function is not supported yet")

            for item in self:
                futures.append(ExecutorPool.submit(f, item, parallel_type=parallel_type))

            futures = self._tqdm(futures, **tqdm_args) if progress_bar else futures
            for future in futures:
                future.result()

    def with_for_each(self, f: Callable[[T], None], progress_bar: bool = False,
        parallel_type: Optional[ParallelType] = None, **tqdm_args
    ) -> Stream[T]:
        self.for_each(f, progress_bar, parallel_type, **tqdm_args)
        return self

    def filter(self, f: Callable[[T], bool]) -> Stream[T]:
        return FilterStream(self, f)

    def reduce(self, f: Callable[[T, T], T]) -> T:
        return reduce(f, self)

    def flatten(self) -> Stream[T]:
        return FlattenStream(self)

    def flat_map(self, f: Callable[[T], Stream[R]]) -> Stream[R]:
        return self.map(f).flatten()

    def skip(self, n: int) -> Stream[T]:
        return SkipStream(self, n)

    def take(self, n: int) -> Stream[T]:
        return TakeStream(self, n)

    def repeat(self, n: int) -> Stream[T]:
        return RepeatStream(self, n)

    @property
    def head(self) -> T:
        return self[0]

    @property
    def tail(self) -> Stream[T]:
        return self.skip(1)

    def to_list(self) -> List[T]:
        return [item for item in self]

    def __str__(self):
        return f"Stream({', '.join(map(str, self._cache + (['...'] if not self._finished else [])))})"


class MapStream(Stream[T]):

    def __init__(self, iterable: Iterable[T], f: Callable[[T], R]):
        super().__init__(iterable)
        self._f = f
        self._cache = []

    def _iter_cond(self, cond: Callable[[int], bool], start: int = 0) -> Generator[T, None, None]:
        i = start
        while cond(i):
            if self._cache_size > i:
                yield self._cache[i]
            elif self._cache_size == i:
                try:
                    item = self._f(next(self._iter))
                except StopIteration:
                    break
                self._cache.append(item)
                yield item
            else:
                for item in self._iter_to(i, start=self._cache_size):
                    yield item
            i += 1


class FilterStream(Stream[T]):

    def __init__(self, iterable: Iterable[T], f: Callable[[T], bool]):
        super().__init__(iterable)
        self._f = f
        self._cache = []

    def _iter_cond(self, cond: Callable[[int], bool], start: int = 0) -> Generator[T, None, None]:
        i = start
        while cond(i):
            if self._cache_size > i:
                yield self._cache[i]
                i += 1
            elif self._cache_size == i:
                try:
                    item = next(self._iter)
                except StopIteration:
                    break
                if self._f(item):
                    self._cache.append(item)
                    yield item
                    i += 1
            else:
                for item in self._iter_to(i, start=self._cache_size):
                    yield item
                    i += 1


class FlattenStream(Stream[T]):

    def __init__(self, iterable: Iterable[T]):
        super().__init__(iterable)
        self._cache = []

    def _iter_cond(self, cond: Callable[[int], bool], start: int = 0) -> Generator[T, None, None]:
        i = start
        while cond(i):
            if self._cache_size > i:
                yield self._cache[i]
                i += 1
            elif self._cache_size == i:
                try:
                    item = next(self._iter)
                except StopIteration:
                    break
                if isinstance(item, Stream):
                    for each in item:
                        self._cache.append(each)
                        yield each
                        i += 1
                else:
                    self._cache.append(item)
                    yield item
                    i += 1
            else:
                for item in self._iter_to(i, start=self._cache_size):
                    yield item
                    i += 1


class TakeStream(Stream[T]):

    def __init__(self, iterable: Iterable[T], n: int):
        super().__init__(iterable)
        self._n = n
        self._cache = []

    def _iter_cond(self, cond: Callable[[int], bool], start: int = 0) -> Generator[T, None, None]:
        def _cond(i: int) -> bool:
            return cond(i) and i < self._n

        return super()._iter_cond(_cond, start=start)


class SkipStream(Stream[T]):

    def __init__(self, iterable: Iterable[T], n: int):
        super().__init__(iterable)
        self._n = n
        self._cache = []
        self._skip_applied = False

    def _apply_skip(self) -> None:
        for _ in range(self._n):
            next(self._iter)
        self._skip_applied = True

    def _iter_cond(self, cond: Callable[[int], bool], start: int = 0) -> Generator[T, None, None]:
        if not self._skip_applied:
            self._apply_skip()
        return super()._iter_cond(cond, start=start)


class RepeatStream(Stream[T]):

    def __init__(self, iterable: Iterable[T], n: int):
        super().__init__(iterable)
        self._cache = []
        self._n = n
        self._iter_once = False

    def _iter_cond(self, cond: Callable[[int], bool], start: int = 0) -> Generator[T, None, None]:
        i = start
        while cond(i):
            if not self._iter_once:
                if self._cache_size > i:
                    yield self._cache[i]
                elif self._cache_size == i:
                    try:
                        item = next(self._iter)
                    except StopIteration:
                        self._iter_once = True
                        continue
                    self._cache.append(item)
                    yield item
                else:
                    for item in self._iter_to(i, start=self._cache_size):
                        yield item

            else:
                if i < len(self._cache) * self._n:
                    yield self._cache[i % self._cache_size]
                else:
                    break
            i += 1



from typing import Dict, Any

import scipy.io


class MatWriter:

    @classmethod
    def to(cls, data: Dict[str, Any], path: str) -> None:
        """
        Write data to a mat file.

        Args:
            data (``Dict[str, Any]``): data to write.
            path (``str``): path to write.
        """
        scipy.io.savemat(path, data)

    def __new__(cls, data: Dict[str, Any], path: str) -> None:
        """Alias to :meth:`~tensorneko_util.io.mat_writer.MatWriter.to`."""
        cls.to(data, path)
